#include "lsup/environment.h"
#include "lsup/data/bootstrap.h"


/*
 * External variables.
 */
LSUP_Term *LSUP_default_ctx = NULL;
LSUP_Buffer *LSUP_default_ctx_buf = NULL;


/**
 * Callbacks.
 */
static int nsmap_comp_fn (const void *a, const void *b, void *udata)
{
    const NSEntry *nsa = a;
    const NSEntry *nsb = b;

    return strncmp (nsa->pfx, nsb->pfx, PFX_LEN);
}

static uint64_t nsmap_hash_fn (
        const void *item, uint64_t seed0, uint64_t seed1)
{
    const NSEntry *nse = item;
    return (uint64_t) LSUP_HASH64 (nse->pfx, strlen (nse->pfx), seed0);
}

static void nsmap_free_fn (void *item)
{ free (((NSEntry *) item)->ns); }


/*
 * API
 */

LSUP_rc
LSUP_init (void)
{
    if (LSUP_IS_INIT) {
        LOG_RC (LSUP_NOACTION);
        return LSUP_NOACTION;
    }

#ifdef DEBUG
    // In debug mode, always use max logging.
    int loglevel = LOG_TRACE;
#else
    char *_loglevel = getenv ("LSUP_LOGLEVEL");
    int loglevel = (_loglevel == NULL) ? LOG_INFO : atoi (_loglevel);
#endif
    log_set_level (loglevel);

    // Default namespace map.
    LSUP_default_nsm = hashmap_new (
            sizeof (NSEntry), 0, LSUP_HASH_SEED, 0,
            nsmap_hash_fn, nsmap_comp_fn, nsmap_free_fn, NULL);

    if (hashmap_oom (LSUP_default_nsm)) return LSUP_MEM_ERR;


    if (UNLIKELY (!LSUP_default_nsm)) return LSUP_ERROR;
    for (int i = 0; init_nsmap[i][0] != NULL; i++)
        LSUP_nsmap_add (init_nsmap[i][0], init_nsmap[i][1]);

    // Default context URI.
    const char *default_ctx_str = getenv ("LSUP_DEFAULT_CTX");
    if (!default_ctx_str) default_ctx_str = DEFAULT_CTX_LABEL;
    LSUP_default_ctx = LSUP_iriref_new (default_ctx_str);
    if (UNLIKELY (!LSUP_default_ctx)) return LSUP_ERROR;
    LSUP_default_ctx_buf = LSUP_term_serialize (LSUP_default_ctx);
    if (UNLIKELY (!LSUP_default_ctx_buf)) return LSUP_ERROR;

    // Initialize term cache.
    LSUP_term_cache = LSUP_term_set_new();
    if (UNLIKELY (!LSUP_term_cache)) return LSUP_MEM_ERR;

    // Create and cache default literal datatype key.
    // This will be done only once in the program, so no need to check for
    // duplicates.
    // This is the last operation that can fail in this function, and it is
    // the indicator for LSUP_IS_INIT.
    LSUP_default_datatype = LSUP_iriref_new (DEFAULT_DTYPE);
    LSUP_rc rc = LSUP_term_set_add (
            LSUP_term_cache, LSUP_default_datatype, NULL);
    PRCCK (rc);

    log_info ("LSUP environment initialized.");

    // Set automatic teardown TODO Is this a good idea?
    atexit (LSUP_done);

    return LSUP_OK;
}


void
LSUP_done (void)
{
    if (!LSUP_IS_INIT) return;

    // Free default NS map and context.
    LSUP_buffer_free (LSUP_default_ctx_buf);
    LSUP_term_free (LSUP_default_ctx);
    hashmap_free (LSUP_default_nsm);

    // Free ID cache, including default literal datatype.
    hashmap_free (LSUP_term_cache);
    LSUP_default_datatype = NULL; // This causes LSUP_IS_INIT to return false.

    log_info ("LSUP environment torn down.");
}


/*
 * Extern inline prototypes.
 */

bool LSUP_is_init (void);
