#include "lsup/store.h"


#define ENTRY(a, b) case LSUP_STORE_##a: return &b;
const LSUP_StoreInt *
LSUP_store_int (LSUP_StoreType type) {
    switch (type) {
        BACKEND_TBL
        default: return NULL;
    }
}
#undef ENTRY


const char *
LSUP_store_type_label (LSUP_StoreType type)
{
#define ENTRY(a, b) case LSUP_STORE_##a: return "STORE_" #a;
    switch (type) {
        BACKEND_TBL
        default: return "";
    }
#undef ENTRY
}


LSUP_Store *
LSUP_store_new (
        const LSUP_StoreType store_type, const char *store_id, size_t size,
        bool clear)
{
    if (UNLIKELY (!LSUP_IS_INIT)) {
        log_error (
                "Environment is not initialized. Did you call LSUP_init()?");
        return NULL;
    }
    const LSUP_StoreInt *sif = LSUP_store_int (store_type);
    if (UNLIKELY (!sif)) {
        log_error ("Not a valid store type: %d", store_type);
        return NULL;
    }

    LSUP_Store *store;
    MALLOC_GUARD (store, NULL);

    store->type = store_type;
    store->sif = sif;
    store->id = store_id ? strdup (store_id) : NULL;

    if (store->sif->setup_fn)
        PCHECK (store->sif->setup_fn (store_id, clear), fail);

    // TODO implement custom default context.
    store->data = store->sif->new_fn (store_id, size);

    return store;

fail:
    free (store->id);
    free (store);

    return NULL;
}


void
LSUP_store_free (LSUP_Store *store)
{
    if (UNLIKELY (!store)) return;

    store->sif->free_fn (store->data);
    if (store->id) free (store->id);
    free (store);
}


size_t
LSUP_store_size (const LSUP_Store *store)
{ return store->sif->size_fn (store->data); }


const LSUP_StoreFeature
LSUP_store_features (const LSUP_Store *store)
{ return store->sif->features; }


char *
LSUP_store_id (const LSUP_Store *store)
{ return store->sif->id_fn (store->data); }


LSUP_rc
LSUP_store_begin (LSUP_Store *store, int flags, void **txn) {
    if (UNLIKELY (!(store->sif->features & LSUP_STORE_TXN)))
        return LSUP_NOT_IMPL_ERR;

    return store->sif->txn_begin_fn (store->data, flags, txn);
}


LSUP_rc
LSUP_store_commit (LSUP_Store *store, void *txn)
{
    if (UNLIKELY (!(store->sif->features & LSUP_STORE_TXN)))
        return LSUP_NOT_IMPL_ERR;

    return store->sif->txn_commit_fn (txn);
}


void
LSUP_store_abort (LSUP_Store *store, void *txn)
{
    if (UNLIKELY (!(store->sif->features & LSUP_STORE_TXN))) return;

    store->sif->txn_abort_fn (txn);
}


LSUP_rc
LSUP_store_update_ctx_txn (
        LSUP_Store *store, void *txn, const LSUP_Buffer *old_c,
        const LSUP_Buffer *new_c)
{
    // TODO Handle txn better for non-txn stores.
    if (UNLIKELY (!(store->sif->features & LSUP_STORE_CTX)))
        return LSUP_NOT_IMPL_ERR;

    return store->sif->update_ctx_fn (store->data, old_c, new_c, txn);
}


void *
LSUP_store_add_init_txn (
        LSUP_Store *store, void *txn, const LSUP_Buffer *sc)
    // TODO Handle txn better for non-txn stores.
{ return store->sif->add_init_fn (store->data, sc, txn); }


LSUP_rc
LSUP_store_add_iter (
        LSUP_Store *store, void *it, const LSUP_BufferTriple *sspo)
{ return store->sif->add_iter_fn (it, sspo); }


void
LSUP_store_add_abort (LSUP_Store *store, void *it)
{
    if (
        UNLIKELY (!(store->sif->features & LSUP_STORE_TXN)) ||
        UNLIKELY (!store->sif->add_abort_fn)
    ) return;

    store->sif->add_abort_fn (it);
}


LSUP_rc
LSUP_store_add_done (LSUP_Store *store, void *it)
{ return store->sif->add_done_fn (it); }


LSUP_rc
LSUP_store_add_term_txn (LSUP_Store *store, void *txn, LSUP_Buffer *sterm)
{ return store->sif->add_term_fn (store, sterm, txn); }


void *
LSUP_store_lookup_txn (
        const LSUP_Store *store, void *txn,
        const LSUP_Buffer *ss, const LSUP_Buffer *sp, const LSUP_Buffer *so,
        const LSUP_Buffer *sc, size_t *ct)
{
    if (UNLIKELY (!(store->sif->features & LSUP_STORE_CTX)) && sc) {
        log_warn (
                "Store does not support contexts. Ignoring passed context.");
        sc = NULL;
    }

    return store->sif->lookup_fn (store->data, ss, sp, so, sc, txn, ct); }


LSUP_rc LSUP_store_iter_next (
        LSUP_Store *store, void *it, LSUP_BufferTriple *sspo,
        LSUP_Buffer **ctx)
{
    if (UNLIKELY (!(store->sif->features & LSUP_STORE_CTX)) && ctx) {
        log_warn (
                "Store does not support contexts. Ignoring passed context.");
        ctx = NULL;
    }

    return store->sif->lu_next_fn (it, sspo, ctx);
}


void
LSUP_store_iter_free (LSUP_Store *store, void *it)
{ store->sif->lu_free_fn (it); }


void *
LSUP_store_iter_txn (const LSUP_Store *store, void *it)
{
    if (UNLIKELY (!(store->sif->features & LSUP_STORE_TXN))) {
        log_warn ("Store does not support transactions.");
        return NULL;
    }

    return store->sif->iter_txn_fn (it);
}


LSUP_rc
LSUP_store_remove_txn (
        LSUP_Store *store, void *txn,
        const LSUP_Buffer *ss, const LSUP_Buffer *sp, const LSUP_Buffer *so,
        const LSUP_Buffer *sc, size_t *ct)
{ return store->sif->remove_fn (store->data, ss, sp, so, sc, txn, ct); }


LSUP_Buffer **
LSUP_store_ctx_list_txn (LSUP_Store *store, void *txn)
{
    if (UNLIKELY (!(store->sif->features & LSUP_STORE_CTX))) {
        log_error ("Store does not support contexts.");
        return NULL;
    }

    return store->sif->ctx_list_fn (store->data, txn);
}
